"""Tools for simulating spatial population distributions."""

import itertools
import multiprocessing
from warnings import warn

import geopandas as gpd
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm.auto import tqdm


def _generate_estimate(input):
    if hasattr(input[0], "_original_data"):
        df = input[0]._original_data.copy()
    else:
        df = input[0].data.copy()
    if input[0].index_type == "singlegroup":
        df = input[1](df, group=input[0].group_pop_var, total=input[0].total_pop_var,)
        estimate = (
            input[0]
            .__class__(df, input[0].group_pop_var, input[0].total_pop_var, **input[2])
            .statistic
        )
    else:
        df = input[1](df, groups=input[0].groups)
        estimate = input[0].__class__(df, input[0].groups, **input[2]).statistic
    return estimate


def simulate_null(
    iterations=500,
    sim_func=None,
    seg_class=None,
    n_jobs=-1,
    backend="loky",
    index_kwargs=None,
):
    """Simulate a series of index values in parallel to serve as a null distribution.

    Parameters
    ----------
    iterations : int, required
        Number of iterations to simulate (size of the distribution), by default 1000
    sim_func : function, required
        population randomization function from segregation.inference to serve as
        the null hypothesis.
    seg_func : Class from segregation.singlegroup or segregation.singlegroup, required
        fitted segregation class from which to generate a reference distribution
    n_jobs : int, optional
        number of cpus to initialize for parallelization. If -1, use all available,
        by default -1
    backend : str, optional
        backend passed to joblib.Parallel, by default "loky"
    index_kwargs : dict, optional
        additional keyword arguments used to fit the index, such as distance or network
        if estimating a spatial index; by default None

    Returns
    -------
    list
        pandas.Series of segregation indices for simulated data
    """
    if not index_kwargs:
        index_kwargs = {}
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()
    estimates = Parallel(n_jobs=n_jobs, backend=backend)(
        delayed(_generate_estimate)((seg_class, sim_func, index_kwargs))
        for i in tqdm(range(iterations))
    )
    return pd.Series(estimates)


def simulate_person_permutation(df, group=None, total=None, groups=None):
    """Simulate the permutation of individuals across spatial units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population data to be randomized
    group : str, optional
        name of column on geodataframe that holds the group total
        (for use with single group indices)
    total : str, optional
        name of column on geodataframe that holds the total population for
        each unit (for use with single group indices)
    groups : list, optional
        list of columns on input dataframe that hold total population counts
        for each group of interest

    Returns
    -------
    geopandas.GeoDataFrame
        geodataframe with randomly reallocated population

    Notes
    -------
    Simulates the random permutation of the existing population's location. Given a pool
    of the total population in the region, randomly allocate each person to a
    geographic unit, subject to the total capacity of each unit. Results are
    guaranteed to respect regional and local totals for geographic units as well
    as regional totals and relative shares for groups
    """
    df = df.copy()
    # ensure we have a coilumn named "index"
    df = df.reset_index(drop=True)
    df = df.reset_index()
    if isinstance(df, gpd.GeoDataFrame):
        geoms = df[[df.geometry.name]]
    else:
        geoms = df.assign(idx=df.index.values)[["idx"]]
    if not total:
        total = "total"
        df["total"] = df[groups].sum(axis=1).astype(int)
    df = df[df[total] > 0]
    if group:
        df[total] = df[total].astype(int)
        df["other"] = df[total] - df[group]
        groups = [group, "other"]

    # create a list of group membership for each person
    members = [[group for i in range(df[group].sum().astype(int))] for group in groups]
    pop_groups = list(itertools.chain.from_iterable(members))

    # create a  list of 1s representing the population in each unit
    df["people"] = df[total].apply(lambda x: [1 for i in range(x)])

    # explode the dataframe to have n_rows = total_population
    df = df["people"].explode().reset_index()["index"].to_frame()
    df["groups"] = pop_groups

    # randomize people's group id
    df["groups"] = df["groups"].sample(frac=1).values

    # reaggregate by unit index
    df = df.groupby("index")["groups"].value_counts().unstack()
    df[total] = df[groups].sum(axis=1)
    df = df.join(geoms, how="right").fillna(0)
    if "idx" in df.columns:
        df = df.drop(columns=["idx"])
        return df

    return gpd.GeoDataFrame(df, geometry=geoms.geometry.name)


def simulate_evenness(df, group=None, total=None, groups=None):
    """Simulate even redistribution of population groups across spatial units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population data to be randomized
    group : str, optional
        name of column on geodataframe that holds the group total
        (for use with single group indices)
    total : str, optional
        name of column on geodataframe that holds the total population for
        each unit (for use with single group indices)
    groups : list, optional
        list of columns on input dataframe that hold total population counts
        for each group of interest

    Returns
    -------
    geopandas.GeoDataFrame
        geodataframe with evenly distributed population groups

    Notes
    -------
    Simulates the random allocation of groups, given the total population of
    each geographic unit (randomizes group totals for each location). Given the total
    population of each location, take draws from a multinomial distribution to assign
    group categories for each person, where the probability of each group is equal to
    its regional share. Results are guaranteed to match local population
    totals, but will include variation in the regional totals for each group
    """
    df = df.copy()
    if df.geometry.name:
        geoms = df[df.geometry.name].values
        crs = df.crs
    if group:
        df[[group, total]] = df[[group, total]].astype(int)
        p_null = df[group].sum() / df[total].sum()

        output = pd.DataFrame()
        output[group] = np.random.binomial(n=df[total].values, p=p_null)
        output[total] = df[total].tolist()
    if groups:
        df = df[groups]
        global_prob_vector = df.sum(axis=0) / df.sum().sum()
        t = df[groups].sum(axis=1).astype(int)

        simul = list(
            map(lambda i: list(np.random.multinomial(i, global_prob_vector)), t)
        )
        output = pd.DataFrame(simul, columns=groups)
    if geoms:
        return gpd.GeoDataFrame(output, geometry=geoms, crs=crs)

    return output


def simulate_systematic_randomization(df, group=None, total=None, groups=None):
    """Simulate systematic redistribution of population groups across spatial units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population data to be randomized
    group : str, optional
        name of column on geodataframe that holds the group total
        (for use with singlegroup indices). 
    total : str, optional
        name of column on geodataframe that holds the total population for
        each unit. For singlegroup indices, this parameter is required. For
        multigroup indices, this is optional if groups are not exhaustive.
    groups : list, optional
        list of columns on input dataframe that hold total population counts
        for each group of interest. Note that if not passing a `total` argument,
        groups are assumed to be exhaustive. If total is not set and groups are not
        exhaustive, the function will estimate incorrect probabilities of choosing
        each geographic unit.

    Returns
    -------
    geopandas.GeoDataFrame
        geodataframe with systematically randomized population groups

    Notes
    -------
    Simulates the random allocation of each group across geographic units, given the total population
    of each group (randomizes location totals for each group). Given the total population of
    each group in the region, take draws from a multinomial distribution where the probability of
    choosing each geographic unit is equal to the total regional share currently residing in the unit.
    Results are guaranteed to respect regional group totals, but will include variation in the total
    population of each geographic unit.

    For more, see Allen, R., Burgess, S., Davidson, R., & Windmeijer, F. (2015). More reliable inference for the dissimilarity index of segregation. The Econometrics Journal, 18(1), 40–66. https://doi.org/10.1111/ectj.12039

    Reference: :cite:`allen2015more`
    """
    if groups:
        if not total:
            warn(
                "No `total` argument passed. Assuming population groups are exhaustive"
            )
            total = "total"
        df[total] = df[groups].sum(axis=1)
    if group:
        assert (
            total
        ), "If simulating a single group, you must also supply a total population column"
        df["other_group_pop"] = df[total] - df[group]
        groups = [group, "other_group_pop"]

    p_j = df[total] / df[total].sum()
    data_aux = {}
    for group in groups:
        n = df[group].sum()
        sim = np.random.multinomial(n, p_j)
        data_aux[group] = sim.tolist()
    df_aux = pd.DataFrame.from_dict(data_aux)
    df_aux[total] = df_aux[groups].sum(axis=1)
    if isinstance(df, gpd.GeoDataFrame):
        df_aux = df[[df.geometry.name]].reset_index().join(df_aux)

    return df_aux


def simulate_bootstrap_resample(df, **kwargs):
    """Generate bootstrap replications of the units with replacement of the same size of the original data.

    Parameters
    ----------
    df : geopandas.GeoDataFrame or pandas.DataFrame
        (geo)dataframe with population counts to be randomized

    Returns
    -------
    DataFrame
        DataFrame with bootstrap resampled observations

    Notes
    -------
    Simulate a synthetic dataset by drawing from rows of the input data with replacement
    until reaching the same number of observations in the original dataframe. Note that if
    input is a geodataframe, then the output will not be planar-enforced, as more than one of
    the same unit may appear in the sample.
    """
    df = df.copy()
    df = df.reset_index(drop=True)
    sample_index = np.random.choice(df.index, size=len(df), replace=True)
    df_aux = df.iloc[sample_index]
    return df_aux


def simulate_geo_permutation(df, **kwargs):
    """Simulate a synthetic dataset by random permutation of geographic units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population counts to be randomized

    Returns
    -------
    DataFrame
        Simulate a synthetic dataset by randomly allocating the units over space
        keeping their original values.
    """
    df = df.copy()
    df = df.reset_index(drop=True)
    data = df.assign(
        geometry=df[df.geometry.name][
            list(np.random.choice(df.shape[0], df.shape[0], replace=False))
        ].reset_index()[df.geometry.name]
    )
    return data


def simulate_systematic_geo_permutation(df, group=None, total=None, groups=None):
    """Simulate systematic redistribution followed by random permutation of geographic units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population data to be randomized
    group : str, optional
        name of column on geodataframe that holds the group total
        (for use with single group indices)
    total : str, optional
        name of column on geodataframe that holds the total population for
        each unit (for use with single group indices)
    groups : list, optional
        list of columns on input dataframe that hold total population counts
        for each group of interest

    Returns
    -------
    geopandas.GeoDataFrame
        geodataframe with systematically randomized population groups
    """
    df = simulate_systematic_randomization(df, group=group, total=total, groups=groups)
    df = simulate_geo_permutation(df)
    return df


def simulate_evenness_geo_permutation(df, group=None, total=None, groups=None):
    """Simulate evenness followed by random permutation of geographic units.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        geodataframe with population data to be randomized
    group : str, optional
        name of column on geodataframe that holds the group total
        (for use with single group indices)
    total : str, optional
        name of column on geodataframe that holds the total population for
        each unit (for use with single group indices)
    groups : list, optional
        list of columns on input dataframe that hold total population counts
        for each group of interest

    Returns
    -------
    geopandas.GeoDataFrame
        geodataframe with evenly distributed population groups
    """
    df = simulate_evenness(df=df, group=group, total=total, groups=groups)
    df = simulate_geo_permutation(df)
    return df


SIMULATORS = {
    "systematic": simulate_systematic_randomization,
    "bootstrap": simulate_bootstrap_resample,
    "evenness": simulate_evenness,
    "person_permutation": simulate_person_permutation,
    "geographic_permutation": simulate_geo_permutation,
    "systematic_permutation": simulate_systematic_geo_permutation,
    "even_permutation": simulate_evenness_geo_permutation,
}



"""Tools for simulating comparative datasets across spatial contexts."""

import numpy as np
import pandas as pd

from .._base import MultiGroupIndex, SingleGroupIndex
from .randomization import simulate_person_permutation


def _prepare_comparative_data(df1, df2, group_pop_var1, group_pop_var2, total_pop_var1, total_pop_var2):
    df1 = df1.copy()
    df2 = df2.copy()
    if hasattr(df1, "geometry"):
        df1 = df1[[group_pop_var1, total_pop_var1, df1.geometry.name]]
    else:
        df1 = df1[[group_pop_var1, total_pop_var1]]

    if hasattr(df2, "geometry"):
        df2 = df2[[group_pop_var2, total_pop_var2, df2.geometry.name]]
    else:
        df2 = df2[[group_pop_var2, total_pop_var2]]

    return df1, df2



def _generate_counterfactual(
    data1,
    data2,
    group_pop_var1,
    total_pop_var1,
    group_pop_var2,
    total_pop_var2,
    counterfactual_approach="composition",
):
    """Generate a counterfactual variables.

    Given two contexts, generate counterfactual distributions for a variable of
    interest by simulating the variable of one context into the spatial
    structure of the other.

    Parameters
    ----------
    data1 : pd.DataFrame or gpd.DataFrame
        Pandas or Geopandas dataframe holding data for context 1

    data2 : pd.DataFrame or gpd.DataFrame
        Pandas or Geopandas dataframe holding data for context 2

    group_pop_var : str
        The name of variable in both data that contains the population size of the group of interest

    total_pop_var : str
        The name of variable in both data that contains the total population of the unit

    approach : str, ["composition", "share", "dual_composition"]
        Which approach to use for generating the counterfactual.
        Options include "composition", "share", or "dual_composition"

    Returns
    -------
    two DataFrames
        df1 and df2  with appended columns 'counterfactual_group_pop', 'counterfactual_total_pop', 'group_composition' and 'counterfactual_composition'

    """
    df1, df2 = DUAL_SIMULATORS[counterfactual_approach](
        data1, data2, group_pop_var1, total_pop_var1, group_pop_var2, total_pop_var2,
    )
    df1["group_composition"] = (df1[group_pop_var1] / df1[total_pop_var1]).fillna(0)
    df2["group_composition"] = (df2[group_pop_var2] / df2[total_pop_var2]).fillna(0)

    df1["counterfactual_composition"] = (
        df1["counterfactual_group_pop"] / df1["counterfactual_total_pop"]
    ).fillna(0)
    df2["counterfactual_composition"] = (
        df2["counterfactual_group_pop"] / df2["counterfactual_total_pop"]
    ).fillna(0)

    df1 = df1.drop(columns=[group_pop_var1, total_pop_var1], axis=1)
    df2 = df2.drop(columns=[group_pop_var2, total_pop_var2], axis=1)

    return df1, df2


def sim_composition(
    df1, df2, group_pop_var1, total_pop_var1, group_pop_var2, total_pop_var2,
):
    """Simulate the spatial distribution of a population group in a region using the CDF of a comparison region.

    For each spatial unit i in region 1, take the unit's percentile in the distribution, and swap the group composition
    with the value of the corresponding percentile in region 2. The composition is the minority population of unit i
    divided by total population of tract i. This approach will shift the relative composition of each spatial
    unit without changing its total population.

    Parameters
    ----------
    df1 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for first dataset with columns holding group and total population counts
    df2 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for second dataset with columns holding group and total population counts
    group_pop_var1 : str
        column holding population counts for group of interest on input df1
    total_pop_var1 : str
        column holding total population counts on input df1
    group_pop_var2 : str
        column holding population counts for group of interest on input df2
    total_pop_var2 : str
        column holding total population counts on input df2

    Returns
    -------
    two pandas.DataFrame
        dataframes with simulated population columns appended
    """
    df1, df2 = _prepare_comparative_data(df1, df2, group_pop_var1, group_pop_var2, total_pop_var1, total_pop_var2)

    df1["group_composition"] = (df1[group_pop_var1] / df1[total_pop_var1]).fillna(0)
    df2["group_composition"] = (df2[group_pop_var2] / df2[total_pop_var2]).fillna(0)

    df1["counterfactual_group_pop"] = (
        df1["group_composition"].rank(pct=True).apply(df2["group_composition"].quantile)
        * df1[total_pop_var1]
    )
    df2["counterfactual_group_pop"] = (
        df2["group_composition"].rank(pct=True).apply(df1["group_composition"].quantile)
        * df2[total_pop_var2]
    )

    df1["counterfactual_total_pop"] = df1[total_pop_var1]
    df2["counterfactual_total_pop"] = df2[total_pop_var2]

    return df1, df2


def sim_dual_composition(
    df1, df2, group_pop_var1, total_pop_var1, group_pop_var2, total_pop_var2,
):
    """Apply the 'composition' for both minority and complementary groups.

    Parameters
    ----------
    df1 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for first dataset with columns holding group and total population counts
    df2 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for second dataset with columns holding group and total population counts
    group_pop_var1 : str
        column holding population counts for group of interest on input df1
    total_pop_var1 : str
        column holding total population counts on input df1
    group_pop_var2 : str
        column holding population counts for group of interest on input df2
    total_pop_var2 : str
        column holding total population counts on input df2

    Returns
    -------
    two pandas.DataFrame
        dataframes with simulated population columns appended

    """
    df1, df2 = _prepare_comparative_data(df1, df2, group_pop_var1, group_pop_var2, total_pop_var1, total_pop_var2)

    df1["group_composition"] = (df1[group_pop_var1] / df1[total_pop_var1]).fillna(0)
    df2["group_composition"] = (df2[group_pop_var2] / df2[total_pop_var2]).fillna(0)

    df1["compl_pop_var"] = df1[total_pop_var1] - df1[group_pop_var1]
    df2["compl_pop_var"] = df2[total_pop_var2] - df2[group_pop_var2]

    df1["compl_composition"] = (df1["compl_pop_var"] / df1[total_pop_var1]).fillna(0)
    df2["compl_composition"] = (df2["compl_pop_var"] / df2[total_pop_var2]).fillna(0)

    df1["counterfactual_group_pop"] = (
        df1["group_composition"].rank(pct=True).apply(df2["group_composition"].quantile)
        * df1[total_pop_var1]
    )
    df2["counterfactual_group_pop"] = (
        df2["group_composition"].rank(pct=True).apply(df1["group_composition"].quantile)
        * df2[total_pop_var2]
    )

    df1["counterfactual_compl_pop"] = (
        df1["compl_composition"].rank(pct=True).apply(df2["compl_composition"].quantile)
        * df1[total_pop_var1]
    )
    df2["counterfactual_compl_pop"] = (
        df2["compl_composition"].rank(pct=True).apply(df1["compl_composition"].quantile)
        * df2[total_pop_var2]
    )

    df1["counterfactual_total_pop"] = (
        df1["counterfactual_group_pop"] + df1["counterfactual_compl_pop"]
    )
    df2["counterfactual_total_pop"] = (
        df2["counterfactual_group_pop"] + df2["counterfactual_compl_pop"]
    )

    return df1, df2


def sim_share(
    df1, df2, group_pop_var1, total_pop_var1, group_pop_var2, total_pop_var2,
):
    """Simulate the spatial population distribution of a region using the CDF of a comparison region.

    For each spatial unit i in region 1, take the unit's percentile in the distribution, and swap the group share
    with the value of the corresponding percentile in region 2. The share is the minority population of unit i
    divided by total population of minority population. This approach will shift the total population of
    each unit without changing the regional proportion of each group

    Parameters
    ----------
    df1 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for first dataset with columns holding group and total population counts
    df2 : pandas.DataFrame or geopandas.GeoDataFrame
        dataframe for second dataset with columns holding group and total population counts
    group_pop_var1 : str
        column holding population counts for group of interest on input df1
    total_pop_var1 : str
        column holding total population counts on input df1
    group_pop_var2 : str
        column holding population counts for group of interest on input df2
    total_pop_var2 : str
        column holding total population counts on input df2

    Returns
    -------
    two pandas.DataFrame
        dataframes with simulated population columns appended

    """
    df1, df2 = _prepare_comparative_data(df1, df2, group_pop_var1, group_pop_var2, total_pop_var1, total_pop_var2)

    df1["compl_pop_var"] = df1[total_pop_var1] - df1[group_pop_var1]
    df2["compl_pop_var"] = df2[total_pop_var2] - df2[group_pop_var2]

    df1["share"] = (df1[group_pop_var1] / df1[group_pop_var1].sum()).fillna(0)
    df2["share"] = (df2[group_pop_var2] / df2[group_pop_var2].sum()).fillna(0)

    df1["compl_share"] = (df1["compl_pop_var"] / df1["compl_pop_var"].sum()).fillna(0)
    df2["compl_share"] = (df2["compl_pop_var"] / df2["compl_pop_var"].sum()).fillna(0)

    # Rescale due to possibility of the summation of the counterfactual share values being grater or lower than 1
    # CT stands for Correction Term
    CT1_2_group = df1["share"].rank(pct=True).apply(df2["share"].quantile).sum()
    CT2_1_group = df2["share"].rank(pct=True).apply(df1["share"].quantile).sum()

    df1["counterfactual_group_pop"] = (
        df1["share"].rank(pct=True).apply(df2["share"].quantile)
        / CT1_2_group
        * df1[group_pop_var1].sum()
    )
    df2["counterfactual_group_pop"] = (
        df2["share"].rank(pct=True).apply(df1["share"].quantile)
        / CT2_1_group
        * df2[group_pop_var2].sum()
    )

    # Rescale due to possibility of the summation of the counterfactual share values being grater or lower than 1
    # CT stands for Correction Term
    CT1_2_compl = (
        df1["compl_share"].rank(pct=True).apply(df2["compl_share"].quantile).sum()
    )
    CT2_1_compl = (
        df2["compl_share"].rank(pct=True).apply(df1["compl_share"].quantile).sum()
    )

    df1["counterfactual_compl_pop"] = (
        df1["compl_share"].rank(pct=True).apply(df2["compl_share"].quantile)
        / CT1_2_compl
        * df1["compl_pop_var"].sum()
    )
    df2["counterfactual_compl_pop"] = (
        df2["compl_share"].rank(pct=True).apply(df1["compl_share"].quantile)
        / CT2_1_compl
        * df2["compl_pop_var"].sum()
    )

    df1["counterfactual_total_pop"] = (
        df1["counterfactual_group_pop"] + df1["counterfactual_compl_pop"]
    )
    df2["counterfactual_total_pop"] = (
        df2["counterfactual_group_pop"] + df2["counterfactual_compl_pop"]
    )
    return df1.fillna(0), df2.fillna(0)


def _prepare_random_label(seg_class_1, seg_class_2):
    if hasattr(seg_class_1, "_original_data"):
        data_1 = seg_class_1._original_data.copy()
    else:
        data_1 = seg_class_1.data.copy()
    if hasattr(seg_class_2, "_original_data"):
        data_2 = seg_class_2._original_data.copy()
    else:
        data_2 = seg_class_2.data.copy()

    data_1["grouping_variable"] = "Group_1"
    data_2["grouping_variable"] = "Group_2"

    if isinstance(seg_class_1, SingleGroupIndex):

        # This step is just to make sure the each frequency column is integer for the approaches and from the same type in order to be able to stack them
        data_1.loc[:, (seg_class_1.group_pop_var, seg_class_1.total_pop_var)] = (
            data_1.loc[:, (seg_class_1.group_pop_var, seg_class_1.total_pop_var)]
            .round(0)
            .astype(int)
        )

        # random permutation needs the columns to have the same names
        data_1 = data_1[
            [seg_class_1.group_pop_var, seg_class_1.total_pop_var, "grouping_variable",]
        ]
        data_1.columns = ["group", "total", "grouping_variable"]

        data_2.loc[:, (seg_class_2.group_pop_var, seg_class_2.total_pop_var)] = (
            data_2.loc[:, (seg_class_2.group_pop_var, seg_class_2.total_pop_var)]
            .round(0)
            .astype(int)
        )
        data_2 = data_2[
            [seg_class_2.group_pop_var, seg_class_2.total_pop_var, "grouping_variable",]
        ]
        data_2.columns = ["group", "total", "grouping_variable"]

        stacked_data = pd.concat([data_1, data_2], axis=0)

    elif isinstance(seg_class_1, MultiGroupIndex):

        groups_list = seg_class_1.groups

        for i in range(len(groups_list)):
            data_1[groups_list[i]] = round(data_1[groups_list[i]]).astype(int)
            data_2[groups_list[i]] = round(data_2[groups_list[i]]).astype(int)

        if seg_class_1.groups != seg_class_2.groups:
            raise ValueError("MultiGroup groups should be the same")

        stacked_data = pd.concat([data_1, data_2], ignore_index=True)
    return stacked_data


def _estimate_random_label_difference(data):
    #  note: if estimating a spatial implicit index, then "space" has already been accounted for...
    #  when the index is computed, the underlying data are transformed to represent the *accessible* population
    #  so when calculating the simulated difference, we need to pop spatial implicit parameters
    
    stacked_data = data[0]
    function = data[1]
    index_args_1 = data[2]
    index_args_2 = data[3]
    idx_type = data[4]
    groups = data[5]
    approach = data[6]
    for args in [index_args_1, index_args_2]:
        if 'network' in args:
            args.pop('network')
        elif 'distance' in args:
            args.pop('distance')

    if approach == 'person_permutation':
        grouping = stacked_data['grouping_variable'].copy().values
        if groups:
            stacked_data = simulate_person_permutation(stacked_data, groups=groups)
        else:
            stacked_data = simulate_person_permutation(stacked_data, group='group', total='total')
        stacked_data['grouping_variable'] = grouping

    else:
        stacked_data["grouping_variable"] = np.random.permutation(
            stacked_data["grouping_variable"].values
        )

    stacked_data_1 = stacked_data[stacked_data["grouping_variable"] == "Group_1"]
    stacked_data_2 = stacked_data[stacked_data["grouping_variable"] == "Group_2"]
    if idx_type == "singlegroup":
        simulations_1 = function(stacked_data_1, "group", "total", **index_args_1)[0]
        simulations_2 = function(stacked_data_2, "group", "total", **index_args_2)[0]
    elif idx_type == "multigroup":
        simulations_1 = function(stacked_data_1, groups, **index_args_1)[0]
        simulations_2 = function(stacked_data_2, groups, **index_args_2)[0]

    est = simulations_1 - simulations_2

    return est


def _estimate_counterfac_difference(data):
    data_1 = data[0]
    data_2 = data[1]
    counterfac_df1 = data[10]
    counterfac_df2 = data[11]

    group_1 = data[2]
    total_1 = data[3]
    group_2 = data[4]
    total_2 = data[5]
    index_args_1 = data[6]
    index_args_2 = data[7]
    approach = data[8]
    function = data[9]

    if approach in ["counterfactual_share", "counterfactual_dual_composition"]:
        data_1[total_1] = counterfac_df1["counterfactual_total_pop"]
        data_2[total_2] = counterfac_df2["counterfactual_total_pop"]

    data_1["fair_coin"] = np.random.uniform(size=len(data_1))
    data_1["test_group_pop_var"] = np.where(
        data_1["fair_coin"] > 0.5,
        data_1[group_1],
        counterfac_df1["counterfactual_group_pop"],
    )

    # Dropping to avoid confusion in the internal function
    data_1_test = data_1.drop([group_1], axis=1)

    simulations_1 = function(
        data_1_test, "test_group_pop_var", total_1, **index_args_1,
    )[0]

    # Dropping to avoid confusion in the next iteration
    data_1 = data_1.drop(["fair_coin", "test_group_pop_var"], axis=1)

    data_2["fair_coin"] = np.random.uniform(size=len(data_2))
    data_2["test_group_pop_var"] = np.where(
        data_2["fair_coin"] > 0.5,
        data_2[group_2],
        counterfac_df2["counterfactual_group_pop"],
    )

    # Dropping to avoid confusion in the internal function
    data_2_test = data_2.drop([group_2], axis=1)

    simulations_2 = function(
        data_2_test, "test_group_pop_var", total_2, **index_args_2,
    )[0]

    # Dropping to avoid confusion in the next iteration
    data_2 = data_2.drop(["fair_coin", "test_group_pop_var"], axis=1)

    est = simulations_1 - simulations_2

    return est


DUAL_SIMULATORS = {
    "composition": sim_composition,
    "dual_composition": sim_dual_composition,
    "share": sim_share,
}



"""Inference wrapper classes for segregation measures."""

__author__ = "Renan X. Cortes <renanc@ucr.edu> Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import multiprocessing
import warnings

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy import stats
from tqdm.auto import tqdm

from .._base import MultiGroupIndex
from .comparative import (DUAL_SIMULATORS, _estimate_counterfac_difference,
                          _estimate_random_label_difference,
                          _generate_counterfactual, _prepare_random_label)
from .randomization import SIMULATORS, simulate_null


def _infer_segregation(
    seg_class,
    iterations_under_null=500,
    null_approach="systematic",
    two_tailed=True,
    index_kwargs=None,
    n_jobs=-1,
    backend="loky",
    null_value=0,
):
    """Compare segregation statistic against a simulated null distribution.

    Parameters
    ----------
    seg_class : segregation.singlegroup or segregation.multigroup object
        fitted segregation index class
    iterations_under_null : int
        number of iterations under null hyphothesis
    null_approach : str
        Which counterfactual approach to use when generating null hypothesis distribution. See Notes.

        * ``systematic``:
        assumes that every group has the same probability with restricted conditional probabilities
        p_0_j = p_1_j = p_j = n_j/n (multinomial distribution).

        * ``bootstrap``:
        generates bootstrap replications of the units with replacement of the same size of the
        original data. This procedure creates a confidence interval for the index statistic to test 
        whether the null value lies within.

        * ``evenness``:
        assumes that each spatial unit has the same global probability of drawing elements from the
        minority group of the fixed total unit population (binomial distribution). 

        * ``person_permutation``:
        randomly allocates individuals into units keeping the total population of each
        equal to the original.

        * ``geographic_permutation``:
        randomly allocates the units over space keeping the original values.

        * ``systematic_permutation``:
        assumes absence of systematic segregation and randomly allocates the units over
        space.

        * ``even_permutation``:
        Assumes the same global probability of drawning elements from the minority group in
        each spatial unit and randomly allocates the units over space.

    two_tailed : boolean
        If True, p_value is two-tailed. Otherwise, it is right one-tailed. The one-tailed p_value attribute
        might not be appropriate for some measures, as the two-tailed. Therefore, it is better to rely on the
        est_sim attribute.
    n_jobs: int, optional
        number of cores to use for estimation. If -1 all available cpus will be used
    backend: str, optional
        which backend to use with joblib. Options include "loky", "multiprocessing", or "threading"
    index_kwargs : dict, optional
        additional keyword arguments passed to the index class

    Attributes
    ----------
    p_value : float
        Pseudo One or Two-Tailed p-value estimated from the simulations
    est_sim : numpy array
       Estimates of the segregation measure under the null hypothesis
    statistic : float
        The value of the segregation index being tested

    """
    if null_approach not in SIMULATORS.keys():
        raise ValueError(f"null_approach must one of {list(SIMULATORS.keys())}")

    if type(two_tailed) is not bool:
        raise TypeError("two_tailed is not a boolean object")

    point_estimation = seg_class.statistic

    # if using the bootstrap test, we're testing the null estimate against the index's distribution
    # in all other cases, we test the index value against a null distribution
    if null_approach == "bootstrap":
        point_estimation = null_value

    aux = str(type(seg_class))
    _class_name = aux[
        1 + aux.rfind(".") : -2
    ]  # 'rfind' finds the last occurence of a pattern in a string

    Estimates_Stars = simulate_null(
        iterations=iterations_under_null,
        sim_func=SIMULATORS[null_approach],
        seg_class=seg_class,
        index_kwargs=index_kwargs,
        n_jobs=n_jobs,
        backend=backend,
    ).values

    # Check and, if the case, remove iterations_under_null that resulted in nan or infinite values
    if any((np.isinf(Estimates_Stars) | np.isnan(Estimates_Stars))):
        warnings.warn(
            "Some estimates resulted in NaN or infinite values for estimations under null hypothesis. "
            "These values will be removed for the final results."
        )
        Estimates_Stars = Estimates_Stars[
            ~(np.isinf(Estimates_Stars) | np.isnan(Estimates_Stars))
        ]

    if not two_tailed:
        p_value = sum(Estimates_Stars > point_estimation) / iterations_under_null
    else:
        aux1 = (point_estimation < Estimates_Stars).sum()
        aux2 = (point_estimation > Estimates_Stars).sum()
        p_value = 2 * np.array([aux1, aux2]).min() / len(Estimates_Stars)

    return p_value, Estimates_Stars, point_estimation, _class_name


class SingleValueTest:
    """Statistical inference for a single segregation measure.

    Parameters
    ----------
    seg_class : segregation.singlegroup or segregation.multigroup object
        fitted segregation index class
    iterations_under_null : int
        number of iterations under null hyphothesis
    null_approach : str
        Which counterfactual approach to use when generating null hypothesis distribution. One of the following:.

        * ``bootstrap``:
        Generate bootstrap replications of the units with replacement of the same size of the
        original data to create a distribution of the segregation index. Then the `null_value` argument
        is tested against this distribution. The null_value may be 0, or may be estimated empirically using
        the `simulate_null` function.

        * ``systematic``:
        assumes that every group has the same probability with restricted conditional probabilities
        p_0_j = p_1_j = p_j = n_j/n (multinomial distribution).

        * ``evenness``:
        Generate a distribution of segregation indices under the assumption of evenness, which
        assumes that each spatial unit has the same global probability of drawing elements from the
        minority group of the fixed total unit population (binomial distribution). Then test the observed
        segregation index against this distribution

        * ``person_permutation``:
        Generate a distribution of segregation indices under the assumption of individual-level randomization,
        which randomly allocates individuals into units keeping the total population of each
        equal to the original.Then test the observed segregation index against this distribution

        * ``geographic_permutation``:
        Generate a distribution of segregation indices under the assumption of geographit unit-level randomization,
        which randomly allocates the units over space keeping the original values. Then test the observed segregation
        index against this distribution

        * ``systematic_permutation``:
        Generate a distribution of segregation indices under the assumption of systemic randomization,
        then randomly allocate units over space. Then test the observed segregation index against this distribution

        * ``even_permutation``:
        Generate a distribution of segregation indices under the assumption of evenness, then randomly allocating
        the units over space. Then test the observed segregation index against this distribution

    two_tailed : boolean
        If True, p_value is two-tailed. Otherwise, it is right one-tailed. The one-tailed p_value attribute
        might not be appropriate for some measures, as the two-tailed. Therefore, it is better to rely on the
        est_sim attribute.
    n_jobs: int, optional
        number of cores to use for estimation. If -1 all available cpus will be used
    backend: str, optional
        which backend to use with joblib. Options include "loky", "multiprocessing", or "threading"
    index_kwargs : dict, optional
        additional keyword arguments passed to the index class

    Attributes
    ----------
    p_value : float
        Pseudo One or Two-Tailed p-value estimated from the simulations
    est_sim : numpy array
       Estimates of the segregation measure under the null hypothesis
    statistic : float
        The value of the segregation index being tested

    Notes
    -----
    1) The different approaches for the null hypothesis affect directly the results of the inference depending on the
    combination of the index type of seg_class and the null_approach chosen. Therefore, the user needs to be aware of
    how these approaches are affecting the data generation process of the simulations in order to draw meaningful
    conclusions. For example, the Modified Dissimilarity (ModifiedDissim) and  Modified Gini (ModifiedGiniSeg) indexes,
    rely exactly on the distance between evenness through sampling which, therefore, the "evenness" value for null
    approach would not be the most appropriate for these indexes.

    Examples
    --------
    Several examples can be found here https://github.com/pysal/segregation/blob/master/notebooks/inference_wrappers_example.ipynb.
    """

    def __init__(
        self,
        seg_class,
        iterations_under_null=500,
        null_approach="systematic",
        two_tailed=True,
        n_jobs=-1,
        **kwargs,
    ):

        aux = _infer_segregation(
            seg_class,
            iterations_under_null,
            null_approach,
            two_tailed,
            n_jobs=n_jobs,
            **kwargs,
        )

        self.p_value = aux[0]
        self.est_sim = aux[1]
        self.statistic = aux[2]
        self._class_name = aux[3]

    def plot(self, color="darkblue", kde=True, ax=None, **kwargs):
        """Plot the distribution of simulated values and the observed index being tested.

        Parameters
        ----------
        color : str, optional
            color of histogram, by default 'darkblue'
        kde : bool, optional
            Whether to plot the kernel density estimate along with the histogram, by default True
        ax : matplotlib.axes, optional
            axes object to plot onto, by default None
        kwargs : seaborn.histplot argument, optional
            additional keyword arguments passed to seaborn's histplot function

        Returns
        -------
        matplotlib.axes
            pyplot axes object
        """
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns
        except ImportError:
            warnings.warn("This method relies on importing `matplotlib` and `seaborn`")

        f = sns.histplot(self.est_sim, color=color, kde=kde, ax=ax, **kwargs)
        plt.axvline(self.statistic, color="red")
        plt.title("{} (Value = {})".format(self._class_name, round(self.statistic, 3)))
        return f


def _compare_segregation(
    seg_class_1,
    seg_class_2,
    iterations,
    null_approach,
    index_kwargs_1,
    index_kwargs_2,
    n_jobs,
    backend,
):
    """Perform inference comparison for a two segregation measures.

    Parameters
    ----------
    seg_class_1 : segregation.singlegroup or segregation.multigroup class
        a fitted segregation class to be compared to seg_class_2
    seg_class_2 : segregation.singlegroup or segregation.multigroup class
        a fitted segregation class to be compared to seg_class_1
    iterations_under_null : int
        number of iterations to simulate observations in a null distribution
    null_approach : str
        Which type of null hypothesis the inference will iterate. One of the following:

        * ``random_label``:
        Randomly assign each spatial unit to a region then recalculate segregation indices and take the difference

        * ``bootstrap``:
        Use bootstrap resampling to generate distributions of the segregation index for each index in the comparison,
        then use a two sample t-test to compare differences in the mean of each distribution

        * ``composition``:
        Generate counterfactual estimates for each region using the sim_composition approach.
        On each iteration, generate a synthetic dataset for each region where each unit has a 50% chance
        of belonging to the original data or the counterfactual data. Recalculate segregation indices on
        the synthetic datasets.

        * ``share``:
        Generate counterfactual estimates for each region using the sim_share approach.
        On each iteration, generate a synthetic dataset for each region where each unit has a 50% chance
        of belonging to the original data or the counterfactual data. Recalculate segregation indices on
        the synthetic datasets.

        * ``dual_composition``:
        Generate counterfactual estimates for each region using the sim_dual_composition
        approach. On each iteration, generate a synthetic dataset for each region where each unit has a 50%
        chance of belonging to the original data or the counterfactual data. Recalculate segregation
        indices on the synthetic datasets.

        * ``person_permutation``:
        Use the simulate_person_permutation approach to randomly reallocate the combined
        population across both regions then recalculate segregation indices

    n_jobs: int, optional
        number of cores to use for estimation. If -1 all available cpus will be used
    backend: str, optional
        which backend to use with joblib. Options include "loky", "multiprocessing", or "threading"
    index_kwargs_1 : dict, optional
        extra parameters to pass to segregation index 1.
    index_kwargs_2 : dict, optional
        extra parameters to pass to segregation index 2.

    Attributes
    ----------
    p_value : float
        Two-Tailed p-value
    est_sim : numpy array
        Estimates of the segregation measure differences under the null hypothesis
    est_point_diff : float
        Observed difference between the segregation measures

    Notes
    -----
    This function performs inference to compare two segregation measures. This can be either two measures of the same locations in two different points in time or it can be two different locations at the same point in time.
    The null hypothesis is H0: Segregation_1 is not different than Segregation_2.

    Based on Rey, Sergio J., and Myrna L. Sastré-Gutiérrez. "Interregional inequality dynamics in Mexico." Spatial Economic Analysis 5.3 (2010): 277-298.

    """
    if not index_kwargs_1:
        index_kwargs_1 = {}
    if not index_kwargs_2:
        index_kwargs_2 = {}
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()
    if null_approach not in [
        "random_label",
        "composition",
        "share",
        "dual_composition",
        "person_permutation",
        "bootstrap",
    ]:
        raise ValueError(
            f"null_approach must one of {list(DUAL_SIMULATORS.keys())+['random_label', 'person_permutation', 'bootstrap']}"
        )

    if type(seg_class_1) != type(seg_class_2):
        raise TypeError("seg_class_1 and seg_class_2 must be the same type/class.")

    point_estimation = seg_class_1.statistic - seg_class_2.statistic

    aux = str(type(seg_class_1))
    _class_name = aux[
        1 + aux.rfind(".") : -2
    ]  # 'rfind' finds the last occurence of a pattern in a string

    data_1 = seg_class_1.data.copy()
    data_2 = seg_class_2.data.copy()

    if null_approach == "bootstrap":
        boot1 = SingleValueTest(
            seg_class_1,
            iterations_under_null=iterations,
            null_approach="bootstrap",
            n_jobs=n_jobs,
            backend=backend,
            **index_kwargs_1,
        ).est_sim
        boot2 = SingleValueTest(
            seg_class_2,
            iterations_under_null=iterations,
            null_approach="bootstrap",
            n_jobs=n_jobs,
            backend=backend,
            **index_kwargs_2,
        ).est_sim
        # test statistic follows from <http://dx.doi.org/10.1016/j.jeconom.2008.11.004>, page 34
        tt = (boot1.mean() - boot2.mean()) / np.sqrt(
            (np.std(boot1) ** 2 + np.std(boot2) ** 2)
        )
        # p-value from <https://docs.scipy.org/doc/scipy/reference/tutorial/stats.html>
        p_value = stats.t.sf(np.abs(tt), iterations - 1) * 2
        estimates = (boot1, boot2)
        return p_value, estimates, point_estimation, _class_name

    if null_approach in ["random_label", "person_permutation"]:
        if isinstance(seg_class_1, MultiGroupIndex):
            groups = seg_class_1.groups
        else:
            groups = None

        stacked = _prepare_random_label(seg_class_1, seg_class_2)

        estimates = Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(_estimate_random_label_difference)(
                (
                    stacked,
                    seg_class_1._function,
                    index_kwargs_1,
                    index_kwargs_2,
                    seg_class_1.index_type,
                    groups,
                    null_approach,
                )
            )
            for i in tqdm(range(iterations))
        )

    if null_approach in [
        "composition",
        "share",
        "dual_composition",
    ]:

        if isinstance(seg_class_1, MultiGroupIndex):
            raise ValueError("Not implemented for MultiGroup indexes.")

        counterfac_df1, counterfac_df2 = _generate_counterfactual(
            data_1,
            data_2,
            seg_class_1.group_pop_var,
            seg_class_1.total_pop_var,
            seg_class_2.group_pop_var,
            seg_class_2.total_pop_var,
            null_approach,
        )

        if null_approach in ["share", "dual_composition"]:
            data_1[seg_class_1.total_pop_var] = counterfac_df1[
                "counterfactual_total_pop"
            ]
            data_2[seg_class_2.total_pop_var] = counterfac_df2[
                "counterfactual_total_pop"
            ]

        estimates = Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(_estimate_counterfac_difference)(
                (
                    data_1,
                    data_2,
                    seg_class_1.group_pop_var,
                    seg_class_1.total_pop_var,
                    seg_class_2.group_pop_var,
                    seg_class_2.total_pop_var,
                    index_kwargs_1,
                    index_kwargs_2,
                    null_approach,
                    seg_class_1._function,
                    counterfac_df1,
                    counterfac_df2,
                )
            )
            for i in tqdm(range(iterations))
        )
    estimates = pd.Series(estimates).dropna()
    if len(estimates) < iterations:
        warnings.warn("Some observations were removed for NA values")

    # Two-Tailed p-value
    # Obs.: the null distribution can be located far from zero. Therefore, this is the the appropriate way to calculate the two tailed p-value.
    aux1 = (point_estimation < estimates).sum()
    aux2 = (point_estimation > estimates).sum()
    p_value = 2 * np.array([aux1, aux2]).min() / len(estimates)

    return p_value, estimates, point_estimation, _class_name


class TwoValueTest:
    """Perform comparative inference for two segregation measures.

    Parameters
    ----------
    seg_class_1 : segregation.singlegroup or segregation.multigroup class
        a fitted segregation class to be compared to seg_class_2
    seg_class_2 : segregation.singlegroup or segregation.multigroup class
        a fitted segregation class to be compared to seg_class_1
    iterations_under_null : int
        number of iterations to simulate observations in a null distribution
    null_approach : str
        Which type of null hypothesis the inference will iterate. One of the following:

        * ``random_label``:
        Randomly assign each spatial unit to a region then recalculate segregation indices and take their
        difference. Repeat this process `iterations` times to generate a reference distribution. Then test
        the observed difference aginst this distribution.

        * ``bootstrap``:
        Use bootstrap resampling to generate distributions of each segregation index in the
        comparison, then use a two sample t-test to compare differences between the distribution means.

        * ``composition``:
        Generate counterfactual estimates for each region using the sim_composition approach.
        On each iteration, generate a synthetic dataset for each region where each unit has a 50% chance
        of belonging to the original data or the counterfactual data. Recalculate segregation indices on
        the synthetic datasets.

        * ``share``:
        Generate counterfactual estimates for each region using the sim_share approach.
        On each iteration, generate a synthetic dataset for each region where each unit has a 50% chance
        of belonging to the original data or the counterfactual data. Recalculate segregation indices on
        the synthetic datasets. Then follow the random labeling method on these synthetic data

        * ``dual_composition``:
        Generate counterfactual estimates for each region using the sim_dual_composition
        approach. On each iteration, generate a synthetic dataset for each region where each unit has a 50%
        chance of belonging to the original data or the counterfactual data. Then follow the random labeling
        method on these synthetic data

        * ``person_permutation``:
        Use the simulate_person_permutation approach to randomly reallocate the combined
        population across both regions then recalculate segregation indices

    n_jobs: int, optional
        number of cores to use for estimation. If -1 all available cpus will be used
    backend: str, optional
        which backend to use with joblib. Options include "loky", "multiprocessing", or "threading"
    index_kwargs_1 : dict, optional
        extra parameters to pass to segregation index 1.
    index_kwargs_2 : dict, optional
        extra parameters to pass to segregation index 2.

    Attributes
    ----------
    p_value : float
        Two-Tailed p-value
    est_sim : numpy array
        Estimates of the segregation measure differences under the null hypothesis
    est_point_diff : float
        Observed difference between the segregation measures

    Notes
    -----
    This function performs inference to compare two segregation measures. This can be either
    two measures of the same locations in two different points in time or it can be two
    different locations at the same point in time. The null hypothesis is H0: Segregation_1
    is not different than Segregation_2.
    Based on Rey, Sergio J., and Myrna L. Sastré-Gutiérrez. "Interregional inequality dynamics in Mexico." Spatial Economic Analysis 5.3 (2010): 277-298.

    Examples
    --------
    Several examples can be found here https://github.com/pysal/segregation/blob/master/notebooks/inference_wrappers_example.ipynb.
    """

    def __init__(
        self,
        seg_class_1,
        seg_class_2,
        iterations_under_null=500,
        null_approach="random_label",
        n_jobs=-1,
        backend="loky",
        index_kwargs_1=None,
        index_kwargs_2=None,
        **kwargs,
    ):

        aux = _compare_segregation(
            seg_class_1,
            seg_class_2,
            iterations=iterations_under_null,
            null_approach=null_approach,
            n_jobs=n_jobs,
            backend=backend,
            index_kwargs_1=index_kwargs_1,
            index_kwargs_2=index_kwargs_2,
        )

        self.p_value = aux[0]
        self.est_sim = aux[1]
        self.est_point_diff = aux[2]
        self._class_name = aux[3]
        self._null_approach = null_approach

    def plot(self, color="darkblue", color2="darkred", kde=True, ax=None, **kwargs):
        """Plot the distribution of simulated values and the index value being tested.

        Parameters
        ----------
        color : str, optional
            histogram color, by default 'darkblue'
        color2: str, optional, by default "darkred"
             Color for second histogram. Only relevant for bootstrap test
        kde : bool, optional
            Whether to plot the kernel density estimate along with the histogram,
            by default True
        ax : matplotlib.axes, optional
            axes object to plot onto, by default None
        kwargs : seaborn.histplot argument, optional
            additional keyword arguments passed to seaborn's histplot function

        Returns
        -------
        matplotlib.axes
            pyplot axes object
        """
        try:
            import matplotlib.pyplot as plt
            import seaborn as sns
        except ImportError:
            warnings.warn("This method relies on importing `matplotlib` and `seaborn`")

        if self._null_approach == "bootstrap":
            ax = sns.histplot(self.est_sim[0], color=color, kde=kde, ax=ax, **kwargs)
            ax = sns.histplot(self.est_sim[1], color=color2, kde=kde, ax=ax, **kwargs)
            plt.title(
                "{} (Diff. value = {})".format(
                    self._class_name, round(self.est_point_diff, 3)
                )
            )
        else:
            ax = sns.histplot(self.est_sim, color=color, kde=kde, ax=ax, **kwargs)
            plt.axvline(self.est_point_diff, color="red")
            plt.title(
                "{} (Diff. value = {})".format(
                    self._class_name, round(self.est_point_diff, 3)
                )
            )
        return ax



"""Relative Concentration Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np

from .._base import SingleGroupIndex, SpatialExplicitIndex


def _relative_concentration(data, group_pop_var, total_pop_var):
    """Calculate Relative Concentration index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit

    Returns
    ----------
    statistic : float
                Relative Concentration Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.


    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.

    """
    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    if any(t < x):
        raise ValueError(
            "Group of interest population must equal or lower than the total population of the units."
        )

    area = np.array(data.area)

    y = t - x

    X = x.sum()
    Y = y.sum()
    T = t.sum()

    # Create the indexes according to the area ordering
    des_ind = (-area).argsort()
    asc_ind = area.argsort()

    # A discussion about the extraction of n1 and n2 can be found in https://github.com/pysal/segregation/issues/43
    n1 = np.where(((np.cumsum(t[asc_ind]) / T) < X / T) == False)[0][0] + 1
    n2_aux = np.where(((np.cumsum(t[des_ind]) / T) < X / T) == False)[0][0] + 1
    n2 = len(data) - n2_aux

    n = data.shape[0]
    T1 = t[asc_ind][0:n1].sum()
    T2 = t[asc_ind][n2:n].sum()

    RCO = (
        (
            ((x[asc_ind] * area[asc_ind] / X).sum())
            / ((y[asc_ind] * area[asc_ind] / Y).sum())
        )
        - 1
    ) / (
        (
            ((t[asc_ind] * area[asc_ind])[0:n1].sum() / T1)
            / ((t[asc_ind] * area[asc_ind])[n2:n].sum() / T2)
        )
        - 1
    )

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return RCO, core_data


class RelativeConcentration(SingleGroupIndex, SpatialExplicitIndex):
    """Relative Concentration Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population

    Attributes
    ----------
    statistic : float
        Relative Conrentration Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        aux = _relative_concentration(
            self.data, self.group_pop_var, self.total_pop_var,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _relative_concentration



"""Boundary Spatial Dissimilarity Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from sklearn.metrics.pairwise import manhattan_distances

from .._base import (SingleGroupIndex, SpatialExplicitIndex,
                     _return_length_weighted_w)
from .dissim import _dissim


def _boundary_spatial_dissim(data, group_pop_var, total_pop_var, standardize=False):
    """Calculation of Boundary Spatial Dissimilarity index.

    Parameters
    ----------
    data : a geopandas DataFrame with a geometry column.
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    standardize : boolean
        A condition for row standardisation of the weights matrices. If True, the values of cij in the formulas gets row standardized.
        For the sake of comparison, the seg R package of Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.
        works by default without row standardization. That is, directly with border length.

    Returns
    ----------
    statistic : float
                Boundary Spatial Dissimilarity Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    The formula is based on Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.

    Original paper by Wong, David WS. "Spatial indices of segregation." Urban studies 30.3 (1993): 559-572.

    References: :cite:`hong2014implementing` and :cite:`wong1993spatial`.

    """
    if type(standardize) is not bool:
        raise TypeError("std is not a boolean object")

    D = _dissim(data, group_pop_var, total_pop_var)[0]

    # If a unit has zero population, the group of interest frequency is zero
    data = data.assign(
        pi=np.where(
            data[total_pop_var] == 0, 0, data[group_pop_var] / data[total_pop_var]
        )
    )

    if not standardize:
        cij = _return_length_weighted_w(data).sparse.todense()
    else:
        cij = _return_length_weighted_w(data).sparse.todense()
        cij = cij / cij.sum(axis=1).reshape((cij.shape[0], 1))

    # manhattan_distances used to compute absolute distances
    num = np.multiply(manhattan_distances(data[["pi"]]), cij).sum()
    den = cij.sum()
    BSD = D - num / den

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return BSD, core_data


class BoundarySpatialDissim(SingleGroupIndex, SpatialExplicitIndex):
    """Boundary-Area Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    standardize : boolean
        A condition for row standardisation of the weights matrices. If True, the values of cij in the formulas gets row standardized.
        For the sake of comparison, the seg R package of Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.
        works by default with row standardization.

    Attributes
    ----------
    statistic : float
        Boundary Area Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    The formula is based on Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.

    Original paper by Wong, David WS. "Spatial indices of segregation." Urban studies 30.3 (1993): 559-572.

    References: :cite:`hong2014implementing` and :cite:`wong1993spatial`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, w=None, standardize=True, **kwargs
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.standardize = standardize
        aux = _boundary_spatial_dissim(
            self.data, self.group_pop_var, self.total_pop_var, self.standardize
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _boundary_spatial_dissim



"""Base classes for segregation indices."""

from warnings import warn
import warnings

import geopandas as gpd
import libpysal
import pandas as pd
from libpysal.weights import lag_spatial
from libpysal.weights.distance import Kernel
from libpysal.weights.util import attach_islands, fill_diagonal

from .network import calc_access
from .util.util import _nan_handle


def _return_length_weighted_w(data):
    """
    Returns a PySAL weights object that the weights represent the length of the common boundary of two areal units that share border.
    Author: Levi Wolf <levi.john.wolf@gmail.com>.
    Thank you, Levi!

    Parameters
    ----------

    data          : a geopandas DataFrame with a 'geometry' column.

    Notes
    -----
    Currently it's not making any projection.

    """

    w = libpysal.weights.Rook.from_dataframe(
        data, ids=data.index.tolist(), geom_col=data._geometry_column_name
    )

    if len(w.islands) == 0:
        w = w
    else:
        warnings.warn("There are some islands in the GeoDataFrame.")
        w_aux = libpysal.weights.KNN.from_dataframe(
            data, ids=data.index.tolist(), geom_col=data._geometry_column_name, k=1
        )
        w = attach_islands(w, w_aux)

    adjlist = w.to_adjlist()
    islands = pd.DataFrame.from_records(
        [{"focal": island, "neighbor": island, "weight": 0} for island in w.islands]
    )
    merged = adjlist.merge(
        data.geometry.to_frame("geometry"),
        left_on="focal",
        right_index=True,
        how="left",
    ).merge(
        data.geometry.to_frame("geometry"),
        left_on="neighbor",
        right_index=True,
        how="left",
        suffixes=("_focal", "_neighbor"),
    )
    # Transforming from pandas to geopandas
    merged = gpd.GeoDataFrame(merged, geometry="geometry_focal")
    merged["geometry_neighbor"] = gpd.GeoSeries(merged.geometry_neighbor)

    # Getting the shared boundaries
    merged["shared_boundary"] = merged.geometry_focal.intersection(
        merged.set_geometry("geometry_neighbor")
    )

    # Putting it back to a matrix
    merged["weight"] = merged.set_geometry("shared_boundary").length
    merged_with_islands = pd.concat((merged, islands))
    length_weighted_w = libpysal.weights.W.from_adjlist(
        merged_with_islands[["focal", "neighbor", "weight"]]
    )
    for island in w.islands:
        length_weighted_w.neighbors[island] = []
        del length_weighted_w.weights[island]

    length_weighted_w._reset()

    return length_weighted_w


class SingleGroupIndex:
    """Class for estimating single-group segregation indices."""

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
    ):
        """Initialize singlegroup index.

        Parameters
        ----------
        data : pandas.DataFrame or geopandas.GeoDataFrame
            dataframe or geodataframe if spatial index holding data for location of interest
        group_pop_var : str
            name of column on dataframe holding population totals for focal group
        total_pop_var : str
            name of column on dataframe holding total overall population
        """

        if any([type(group_pop_var) is not str, type(total_pop_var) is not str]):
            raise TypeError("group_pop_var and total_pop_var must be strings")

        if any([group_pop_var not in data.columns, total_pop_var not in data.columns]):
            raise ValueError(
                "group_pop_var and total_pop_var columns must be present on the dataframe"
            )

        if any(data[total_pop_var] < data[group_pop_var]):
            raise ValueError(
                "Group of interest population must equal or lower than the total population of the units."
            )

        if isinstance(data, gpd.GeoDataFrame):
            data = _nan_handle(
                data[[group_pop_var, total_pop_var, data.geometry.name]]
            )
            data = data[[group_pop_var, total_pop_var, data.geometry.name]]

        else:
            data = _nan_handle(data[[group_pop_var, total_pop_var]])
            data =data[[group_pop_var, total_pop_var]]

        data["group_2_pop_var"] = data[total_pop_var] - data[group_pop_var]

        self.index_type = "singlegroup"
        self.data = data.copy()
        self.group_pop_var = group_pop_var
        self.total_pop_var = total_pop_var


class MultiGroupIndex:
    """Class for estimating multi-group segregation indices."""

    def __init__(
        self,
        data,
        groups,
    ):
        """Initialize multi-group index.

        Parameters
        ----------
        data : pandas.DataFrame or geopandas.GeoDataFrame
            dataframe or geodataframe if spatial index holding data for location of interest
        groups : list-like
            list of column names on input DataFrame that hold population totals for groups of interest
        """
        self.index_type = "multigroup"
        self.data = data.copy()
        self.groups = groups


class SpatialExplicitIndex:
    """Class for estimating segregation indices that are explicitly spatial (have no aspatial version)."""

    def __init__(self, **kwargs):
        """Initialize spatially explicit index."""
        if not isinstance(self.data, gpd.GeoDataFrame):
            raise TypeError(
                "`data` must be a geopanads.GeoDataFrame with a vaild geometry column"
            )
        if self.data.crs.is_geographic:
            warn(
                "Geometry is in a geographic CRS. Distance and area calculations in this index are likely incorrect. "
                "Re-project the input data to a projected CRS using `GeoDataFrame.to_crs()` before calculating this index."
            )
        self.spatial_type = "explicit"


class SpatialImplicitIndex:
    """Class for estimating segregation indices that can be spatial or aspatial."""

    def __init__(
        self,
        w,
        network,
        distance=1000,
        decay="linear",
        function="triangular",
        precompute=False,
        **kwargs
    ):
        """Initialize spatially implicit index.

        Parameters
        ----------
        w : libpysal.weights object
            lipysal spatial kernel weights object used to define an egohood
        network : pandana.Network
            pandana Network object representing the study area
        distance : int
            Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
        decay : str
            type of decay function to apply (passed to `pandana`). Options include
            {'linear', 'exponential', or 'flat'}
        function : str
            decay function to use in spatial weights object (passed to libpysal.weights.Kernel)
            options = {'triangular','uniform','quadratic','quartic','gaussian'}
        precompute : bool
            Whether to precompute the pandana Network object
        """
        self.spatial_type = "implicit"
        if self.index_type == "multigroup":
            self._groups = self.groups
        elif self.index_type == "singlegroup":
            self._groups = [self.group_pop_var, self.total_pop_var, "group_2_pop_var"]

        if w and network:
            raise AttributeError(
                "must pass either a pandana network or a pysal weights object, but not both"
            )
        if network:
            access = calc_access(
                self.data,
                variables=self._groups,
                network=network,
                distance=distance,
                decay=decay,
                precompute=precompute,
            )
            self._original_data = self.data.copy()
            self.data = access
            self.network = network
        elif w:
            self.data = _build_local_environment(
                self.data, self._groups, w, function=function
            )
            self.w = w
        elif distance and not network:
            self._original_data = self.data.copy()
            self.data = _build_local_environment(
                self.data, self._groups, bandwidth=distance, function=function
            )


def _build_local_environment(
    data, groups, w=None, bandwidth=1000, function="triangular"
):
    """Convert observations into spatially-weighted sums.

    Parameters
    ----------
    data : GeoDataFrame
        dataframe with local observations
    w : libpysal.weights object
        weights matrix defining the local environment

    Returns
    -------
    DataFrame
        Spatialized data

    """
    data = data.copy().reset_index()
    if data.crs.is_geographic:
        warnings.warn(
            "GeoDataFrame appears to have a geographic coordinate system and likely needs to be reprojected"
        )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if not w:
            w = Kernel.from_dataframe(data, bandwidth=bandwidth, function=function)
        new_data = []
        w = fill_diagonal(w)
        for y in data[groups]:
            new_data.append(lag_spatial(w, data[y]))
        new_data = pd.DataFrame(dict(zip(groups, new_data))).round(0)
        new_data = data.geometry.to_frame().join(new_data.reset_index())

        return new_data



"""Multigroup Local Simpson Interaction index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_local_simpson_interaction(data, groups):
    """
    Calculation of Local Simpson Interaction index for each unit

    Parameters
    ----------

    data   : a pandas DataFrame of n rows
    
    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistics : np.array(n)
                 Local Simpson Interaction values for each unit
                
    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on the local version of Equation 1 of page 37 of Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.
    
    Simpson's interaction index can be simply interpreted as the probability that two individuals chosen at random and independently from the population will be found to not belong to the same group.

    Higher values means lesser segregation.
    
    Simpson's Concentration + Simpson's Interaction = 1
    
    Reference: :cite:`reardon2002measures`.

    """

    core_data = data[groups]

    df = np.array(core_data)

    ti = df.sum(axis=1)
    pik = df / ti[:, None]

    local_SI = np.nansum(pik * (1 - pik), axis=1)

    return local_SI, core_data, groups


class MultiLocalSimpsonInteraction(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Local Simpson Interaction Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular'
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_local_simpson_interaction(self.data, self.groups)

        self.statistics = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_local_simpson_interaction



"""Multigroup Relative Diversity index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from geopandas import GeoDataFrame

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_relative_diversity(data, groups):
    """
    Calculation of Multigroup Relative Diversity index

    Parameters
    ----------

    data   : a pandas DataFrame

    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistic : float
                Multigroup Relative Diversity Index

    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F. "Measures of racial diversity and segregation in multigroup and hierarchically structured populations." annual meeting of the Eastern Sociological Society, Philadelphia, PA. 1998.

    High diversity means less segregation.

    Reference: :cite:`reardon1998measures`.

    """

    core_data = data[groups]
    df = np.array(core_data)

    T = df.sum()

    ti = df.sum(axis=1)
    pik = df / ti[:, None]
    pik = np.nan_to_num(pik)  # Replace NaN from zerodivision when unit has no population
    Pk = df.sum(axis=0) / df.sum()
    Is = (Pk * (1 - Pk)).sum()

    MRD = (ti[:, None] * (pik - Pk) ** 2).sum() / (T * Is)
    if isinstance(data, GeoDataFrame):
        core_data = data[[data.geometry.name]].join(core_data)
    return MRD, core_data, groups


class MultiRelativeDiversity(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Relative Diversity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.

    Reference: :cite:`reardon2002measures`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular',
        **kwargs
    ):
        """Init."""
        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_relative_diversity(self.data, self.groups)

        self.statistic = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_relative_diversity



from .absolute_centralization import AbsoluteCentralization
from .absolute_clustering import AbsoluteClustering
from .absolute_concentration import AbsoluteConcentration
from .atkinson import Atkinson
from .bias_corrected_dissim import BiasCorrectedDissim
from .boundary_spatial_dissim import BoundarySpatialDissim
from .conprof import ConProf
from .correlationr import CorrelationR
from .delta import Delta
from .density_corrected_dissim import DensityCorrectedDissim
from .dissim import Dissim
from .dx_interaction import DistanceDecayInteraction
from .dx_isolation import DistanceDecayIsolation
from .entropy import Entropy
from .gini import Gini
from .interaction import Interaction
from .isolation import Isolation
from .minmax import MinMax
from .modified_dissim import ModifiedDissim
from .modified_gini import ModifiedGini
from .par_dissim import PARDissim
from .relative_centralization import RelativeCentralization
from .relative_clustering import RelativeClustering
from .relative_concentration import RelativeConcentration
from .spatial_dissim import SpatialDissim
from .spatial_prox_profile import SpatialProxProf
from .spatial_proximity import SpatialProximity



"""ConProf Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np
import pandas as pd

from .._base import SingleGroupIndex, SpatialImplicitIndex


def _conprof(data, group_pop_var, total_pop_var, m=1000):
    """Calculation of Concentration Profile.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame
        Dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : string
        Variable containing the population count of the group of interest
    total_pop_var : string
        Variable in data that contains the total population count of the unit

    Returns
    ----------
    statistic : float
        MinMax index statistic value
    core_data : pandas.DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Hong, Seong-Yun, and Yukio Sadahiro. "Measuring geographic segregation: a graph-based approach." Journal of Geographical Systems 16.2 (2014): 211-231.

    Reference: :cite:`hong2014measuring`.

    """
    if type(m) is not int:
        raise TypeError("m must be a string.")

    if m < 2:
        raise ValueError("m must be greater than 1.")

    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    if any(t < x):
        raise ValueError(
            "Group of interest population must equal or lower than the total population of the units."
        )

    def calculate_vt(th):
        g_t_i = np.where(x / t >= th, 1, 0)
        v_t = (g_t_i * x).sum() / x.sum()
        return v_t

    grid = np.linspace(0, 1, m)
    curve = np.array(list(map(calculate_vt, grid)))

    threshold = x.sum() / t.sum()
    R = (
        threshold
        - ((curve[grid < threshold]).sum() / m - (curve[grid >= threshold]).sum() / m)
    ) / (1 - threshold)

    return R, grid, curve, data


class ConProf(SingleGroupIndex, SpatialImplicitIndex):
    """ConProf Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
                ConProf Index
    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Hong, Seong-Yun, and Yukio Sadahiro. "Measuring geographic segregation: a graph-based approach." Journal of Geographical Systems 16.2 (2014): 211-231.

    Reference: :cite:`hong2014measuring`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        w=None,
        network=None,
        distance=None,
        decay=None,
        function="triangular",
        precompute=None,
        **kwargs
    ):
        """Init."""

        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        aux = _conprof(self.data, self.group_pop_var, self.total_pop_var)

        self.statistic = aux[0]
        self.grid = aux[1]
        self.curve = aux[2]
        self.core_data = aux[3]
        self._function = _conprof

    def plot(self):
        """Plot the Concentration Profile."""
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("plotting requires `matplotlib`")
        graph = plt.scatter(self.grid, self.curve, s=0.1)
        return graph



from .local_multi_diversity import MultiLocalDiversity
from .local_multi_entropy import MultiLocalEntropy
from .local_multi_location_quotient import MultiLocationQuotient
from .local_multi_simpson_concentration import MultiLocalSimpsonConcentration
from .local_multi_simpsons_interaction import MultiLocalSimpsonInteraction
from .local_relative_centralization import LocalRelativeCentralization
from .local_distortion import LocalDistortion



"""Multigroup dissimilarity index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_local_entropy(data, groups):
    """
    Calculation of Local Entropy index for each unit

    Parameters
    ----------

    data   : a pandas DataFrame of n rows
    
    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistics : np.array(n)
                 Local Entropy values for each unit
                
    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Eq. 6 of pg. 139 (individual unit case) of Reardon, Sean F., and David O’Sullivan. "Measures of spatial segregation." Sociological methodology 34.1 (2004): 121-162.
    
    Reference: :cite:`reardon2004measures`.

    """

    core_data = data[groups]

    df = np.array(core_data)

    K = df.shape[1]

    ti = df.sum(axis=1)
    pik = df / ti[:, None]

    multi_LE = -np.nansum((pik * np.log(pik)) / np.log(K), axis=1)

    return multi_LE, core_data, groups


class MultiLocalEntropy(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Local Entropy Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Eq. 6 of pg. 139 (individual unit case) of Reardon, Sean F., and David O’Sullivan. "Measures of spatial segregation." Sociological methodology 34.1 (2004): 121-162.
    
    Reference: :cite:`reardon2004measures`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular'
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_local_entropy(self.data, self.groups)

        self.statistics = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_local_entropy



"""Multigroup dissimilarity index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from geopandas import GeoDataFrame

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_dissim(data, groups):
    """Calculation of Multigroup Dissimilarity index.

    Parameters
    ----------
    data : pandas.DataFrame
        DataFrame holding counts of population groups
    groups : list of strings.
        The variables names in data of the groups of interest of the analysis.

    Returns
    -------
    statistic : float
        Multigroup Dissimilarity Index
    core_data : pandas.DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Sakoda, James M. "A generalized index of dissimilarity." Demography 18.2 (1981): 245-250.

    Reference: :cite:`sakoda1981generalized`.

    """
    core_data = data[groups]
    df = np.array(core_data)

    n = df.shape[0]
    K = df.shape[1]

    T = df.sum()

    ti = df.sum(axis=1)
    pik = df / ti[:, None]
    pik = np.nan_to_num(pik)  # Replace NaN from zerodivision when unit has no population
    Pk = df.sum(axis=0) / df.sum()

    Is = (Pk * (1 - Pk)).sum()

    multi_D = (
        1
        / (2 * T * Is)
        * np.multiply(abs(pik - Pk), np.repeat(ti, K, axis=0).reshape(n, K)).sum()
    )
    if isinstance(data, GeoDataFrame):
        core_data = data[[data.geometry.name]].join(core_data)
    return multi_D, core_data, groups


class MultiDissim(MultiGroupIndex, SpatialImplicitIndex):
    """Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular',
        **kwargs
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_dissim(self.data, self.groups)

        self.statistic = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_dissim



import geopandas as gpd

from .._base import MultiGroupIndex, SpatialExplicitIndex
from ..dynamics import compute_divergence_profiles


def _local_distortion(
    gdf, groups, metric="euclidean", network=None, distance_matrix=None, normalize=False
):
    """
    A segregation metric, using Kullback-Leiber (KL) divergence to quantify the
    difference in the population characteristics between (1) an area and (2) the total population.

    This function utilises the methodology proposed in
    Olteanu et al. (2019): 'Segregation through the multiscalar lens'. Which can be
    found here: https://doi.org/10.1073/pnas.1900192116

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    metric : str (optional; 'euclidean' by default)
        Distance metric for calculating pairwise distances,
        Accepts any inputs to `scipy.spatial.distance.pdist`.
        Ignored if passing a network or distance matrix
    network: pandana.Network object (optional, None by default)
        A pandana Network object used to compute distance between observations
    distance_matrix: numpy.array (optional; None by default)
        numpy array of distances between observations in the dataset
    normalize: bool
        NOT YET IMPLEMENTED


    Returns
    ----------
    aux : geopandas.GeoDataFrame
        geodataframe of distortion coefficient values

    """
    # Store the observation index to return with the results
    geoms = gdf[gdf.geometry.name]
    centroids = gdf.geometry.centroid

    aux = compute_divergence_profiles(
        gdf=gdf,
        groups=groups,
        network=network,
        metric=metric,
        distance_matrix=distance_matrix,
    )
    # divergence --> distortion by summing at each location
    aux = gpd.GeoDataFrame(
        aux.groupby("observation").sum()[["divergence"]], geometry=geoms
    ).rename(columns={"divergence": "distortion"})
    if normalize:
        raise Exception("Not yet implemented")
        # Need to write a routine to determine the scaling factor... From the paper:

        # the maximum distortion coefficient in a theoretical extreme case of segregation.
        # Theoretically, the maximal-segregation distortion coefficient is achieved when sorting
        # the k groups into k ghettos, ordered by sizes, and then computing the coefficient for
        # the most isolated person in the smallest group

    return aux


class LocalDistortion(MultiGroupIndex, SpatialExplicitIndex):
    """Multigroup Local Distortion Coefficients.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    metric : str (optional; 'euclidean' by default)
        Distance metric for calculating pairwise distances,
        Accepts any inputs to `scipy.spatial.distance.pdist`.
        Ignored if passing a network or distance matrix
    network: pandana.Network object (optional; None by default)
        A pandana Network object used to compute distance between observations
    distance_matrix:
        numpy array of distances between observations in the dataset
    normalization:
        NOT YET IMPLEMENTED

    Attributes
    ----------
    statistics : pandas.Series
        KL Divergence coefficients
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Olteanu et al. (2019): 'Segregation through the multiscalar lens'.  https://doi.org/10.1073/pnas.1900192116

    """

    def __init__(
        self,
        data,
        groups=None,
        metric="euclidean",
        network=None,
        distance_matrix=None,
        normalize=False,
        **kwargs
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        SpatialExplicitIndex.__init__(self)

        aux = _local_distortion(
            self.data,
            self.groups,
            network=network,
            metric=metric,
            normalize=normalize,
            distance_matrix=distance_matrix,
        )

        self.statistics = aux["distortion"]
        self.data = aux
        self._function = _local_distortion
